'''
Created on Apr 14, 2011

@author: Panagiotis

Usage: visuzalize [options] <bids> <conflicts> <winners file>

Options
    --beta, -b           bid increase for each conflict as a proportion of the original bid
    --output=, -o        output directory
    --format=, -f        output format: png (default), ps  
    --help, -h           help
    --auctions, -a       comma separated list of keyword auctions. Default: all
    --suffix, -s         suffix of the output file (excluding extension)
  
'''

import sys
import logging
import getopt
import pydot
import os.path

from collections import defaultdict
from operator import itemgetter


def main(argv=None):
  if argv is None:
    argv = sys.argv
  
  try:
    opts, args = getopt.getopt(argv[1:], 
                               'b:ha:o:f:s:', 
                               ['beta=',  
                                'help', 
                                'auctions=',
                                'output=',
                                'format=',
                                'suffix='])
  except getopt.GetoptError as e:
    print str(e)
    print __doc__
    return 1
  if len(args) != 3:
    print __doc__
    return 1
  bids_fn = args[0]
  conflicts_fn = args[1]
  results_fn = args[2]
  output_dir = '.'
  format = 'png'
  beta = 0
  suffix = ''
  for o, v in opts:
    if o in ('--help', '-h'):
      print __doc__
      return 0
    elif o in ('--auctions', '-a'):
      keywords = map(int, v.split(','))
    elif o in ('--beta', '-b'):
      beta = float(v)
      if beta < 0 or beta > 1:
        return 1
    elif o in ('--output', '-o'):
      output_dir = v
    elif o in ('--format', '-f'):
      if v in ('png', 'ps', 'pdf'):
        format = v
      else:
        print __doc__
        return 1
    elif o in ('--suffix', '-s'):
      suffix = v + '.'
      
  logging.basicConfig(format="%(asctime)s\t%(levelname)s\t%(message)s", level=logging.INFO)
  logging.info('Reading bids...')
  keyword2bids = read_bids(bids_fn)
  logging.info('... done.')
  logging.info('Reading conflicts...')
  advertiser2conflicts = read_conflicts(conflicts_fn)
  logging.info('... done.')
  if beta > 0:
    adjust_bids(keyword2bids, advertiser2conflicts, beta)
  
  if keywords is None:
    keywords = sorted(keyword2bids.iterkeys())
  else:
    keywords.sort()
  
  cnt = 0
  for keyword, (results, stats) in zip(keywords, read_results(results_fn, keywords)):
    cnt += 1
    # Fetch stats
    graph = construct_graph(keyword2bids[keyword], advertiser2conflicts, results, stats)
    out_fn = os.path.join(output_dir, '%d.%s%s' % (keyword, suffix, format))
    graph.write(out_fn, format=format)
    logging.info('File %s was created.' % (out_fn))
 

GRAY_COLORS = ["#4C4C4C", "#6C6C6C", "#838383", "#969696",
               "#A7A7A7", "#B5B5B5", "#C3C3C3", "#CFCFCF",
               "#DBDBDB", "#E6E6E6"]  
def construct_graph(advertiser2bid, advertiser2conflicts, results, stats=''):
  graph = pydot.Dot(graph_type='digraph', label=stats, fontsize=36)
  graph.set_node_defaults(fontsize=24)
  graph.set_edge_defaults(style='setlinewidth(1.5)')
  graph.set_ranksep = 3
  graph.set_nodesep = 0.25
  
  
  nodes = set() 
  # Add winners
  color_index = 0
  last_bid = None
  for advertiser, price in results:
    color = GRAY_COLORS[color_index]
    bid = advertiser2bid[advertiser]
    node = pydot.Node(str(advertiser), style='filled', fillcolor=color, 
                      label='$%.2f\n($%.2f)' % (bid, price))
    graph.add_node(node)
    nodes.add(advertiser)
    if bid != last_bid:
      color_index += 1
      last_bid = bid
  winners = set(nodes)
  # Add winner conflicts
  for winner in winners:
    conflicts = []
    for conflict in advertiser2conflicts[winner]:
      if conflict in advertiser2bid:
        conflicts.append((conflict, advertiser2bid[conflict]))
    for conflict, value in sorted(conflicts, key=itemgetter(1), reverse=True)[:3]:
      if conflict not in nodes:
        graph.add_node(pydot.Node(str(conflict), label='$%.2f' % (advertiser2bid[conflict])))
        nodes.add(conflict)
  # Add top-cnt nodes not in the graph
  cnt = 3
  for advertiser, bid in sorted(advertiser2bid.iteritems(), key=itemgetter(1), reverse=True):
    if advertiser not in nodes:
      graph.add_node(pydot.Node(str(advertiser), label='$%.2f' % (advertiser2bid[advertiser])))
      nodes.add(advertiser)
      cnt -= 1
      if cnt == 0:
        break
  # Add edges
  for advertiser in nodes:
    for conflict in (nodes & advertiser2conflicts[advertiser]):
      graph.add_edge(pydot.Edge(str(advertiser), str(conflict)))
  
  return graph
    
def adjust_bids(keyword2bids, advertiser2conflicts, beta):
  for keyword, advertiser2bid in keyword2bids.iteritems():
    for advertiser, bid in advertiser2bid.iteritems():
      n = len(advertiser2conflicts[advertiser]) 
      if n > 0:
        advertiser2bid[advertiser] += n * beta * bid
     

def read_bids(bids_fn):
  bids_f = open(bids_fn)
  bids_f.readline()                      # skip header
  keyword2bids = defaultdict(dict)
  for line in bids_f:
    ws = line.strip().split()
    advertiser = int(ws[0])
    keyword = int(ws[1])
    bid = float(ws[2])
    keyword2bids[keyword][advertiser] = bid
  bids_f.close()
  return keyword2bids

def read_conflicts(conflicts_fn):
  advertiser2conflicts = defaultdict(set)
  for line in open(conflicts_fn):
    ws = map(int, line.split()[0:2])
    
    advertiser2conflicts[ws[0]].add(ws[1])
    #advertiser2conflicts[ws[1]].add(ws[0])
  return advertiser2conflicts

def read_results(results_fn, keywords=None):
  keyword_index = 0
  for line in open(results_fn):
    if keywords is not None and keyword_index >= len(keywords):
      return
    ws = line.strip().split(' ', 1)
    keyword = int(ws[0])
    if keywords is None or keyword == keywords[keyword_index]:
      ws = line.strip().split(' ')
      results = []
      i = 10
      while i + 2 < len(ws):
        advertiser = int(ws[i])
        price = float(ws[i + 2])
        results.append((advertiser, price))
        i += 3
      keyword_index += 1
      yield results, '#advertisers: %d, #conflicts: %d, time: %.0f msec' % (int(ws[3]), int(ws[5]), 1000 * float(ws[6]))

if __name__ == '__main__':
  sys.exit(main(sys.argv))
