'''
Created on Apr 14, 2011

@author: Panagiotis

Usage: winners [options] <bids> <conflicts>

Options 
    --beta, -b           bid increase for each conflict as a proportion of the original bid
    --help, -h           help
    --auctions, -a       comma separated list of keyword auctions. Default: all
    --k, -k              number of winner slots
    --logging=, -l       logging level: warn (default), info, debug
    --results=, -r       outuput results: 'a' aggregate (default), 'l' local, 'b' both
    --m=, -m             number of slots
    --sigma=, -s         exponent of the zipfian distribution for the click probabilities
    --winner=, -w        winner determination algorithm: 'debbassac' (default), 'ca'
  
'''
import sys
import logging
import getopt
from collections import defaultdict

from determination.debbassac import Debbassac 
from determination.combinatorialauction import CombinatorialAuction

def main(argv=None):
  if argv is None:
    argv = sys.argv
  
  try:
    opts, args = getopt.getopt(argv[1:], 
                               'b:ha:m:l:r:s:w:', 
                               ['beta=',  
                                'help', 
                                'auctions=', 
                                'm=', 
                                'logging=',
                                'results=',
                                'sigma=',
                                'winner='])
  except getopt.GetoptError as e:
    print str(e)
    print __doc__
    return 1
  if len(args) != 2:
    print __doc__
    return 1
  bids_fn = args[0]
  conflicts_fn = args[1]
  use_local = False
  m = 10
  keywords = None
  greedy = 1
  beta = 0
  to_output_local = False
  to_output_aggregate = True
  sigma = 0.5
  allocator_class = Debbassac
  LOGGING_LEVELS = {'warn': logging.WARN, 
                    'info': logging.INFO,
                    'debug': logging.DEBUG}
  logging_level = logging.INFO
  for o, v in opts:
    if o in ['--help', '-h']:
      print __doc__
      return 0
    elif o in ['--auctions', '-a']:
      keywords = map(int, v.split(','))
    elif o in ['--m', '-m']:
      k = int(v)
    elif o in ['--logging', '-l']:
      if v not in LOGGING_LEVELS:
        print __doc__
        return 1
      logging_level = LOGGING_LEVELS[v] 
    elif o in ['--greedy', '-g']:
      greedy = int(v)
      if greedy not in [1, 2]:
        print __doc__
        return 1
    elif o in ['--beta', '-b']:
      beta = float(v)
      if beta < 0 or beta > 1:
        return 1
    elif o in ['--results', '-r']:
      if v not in ['a', 'l', 'b']:
        print __doc__
        return 1
      to_output_local = True
      to_output_aggregate = True
      if v == 'a':
        to_output_local = False
      elif v == 'l':
        to_output_aggregate = False
    elif o in ['--sigma', '-s']:
      sigma = float(v)
    elif o in ['--winner', '-w']:
      v = v.lower()
      if v == 'debbassac':
        allocator_class = Debbassac
      elif v == 'ca':
        allocator_class = CombinatorialAuction
      else:
        print __doc__
        return 1
  logging.basicConfig(format="%(asctime)s\t%(levelname)s\t%(message)s", level=logging_level)
  
  keyword2bids = read_bids(bids_fn)
  advertiser2conflicts = read_conflicts(conflicts_fn)
  adjust_bids(keyword2bids, advertiser2conflicts, beta)
  
  if keywords is None:
    keywords = keyword2bids.iterkeys()
  
  rhos = [1.0 / (i**sigma) for i in range(1, m + 1)] 
  cnt = 0
  sum_bids = 0
  sum_welfares = 0
  sum_prices = 0
  sum_revenues = 0
  sum_tx = 0
  sum_tp = 0
  for keyword in keywords:
    #logging.warn('Keyword=%d' % keyword)
    logging.info('Keyword: %d' % (keyword))
    #configuration = determine_winners(keyword2bids[keyword], advertiser2conflicts, k, use_local=use_local, greedy=greedy)
    allocator = allocator_class(keyword2bids[keyword], advertiser2conflicts, rhos, keyword)
    allocator.allocate_slots(to_price=True)
    cnt += 1
    # Fetch stats
    stats = {'n': allocator.n,
             'm': allocator.m,
             'd': allocator.d,
             'tx': allocator.delta_winners,
             'tp': allocator.delta_price,
             'w': allocator.w_star,
             'r': sum([p * rho for p, rho in 
                       zip(allocator.p_star, 
                           rhos[:len(allocator.p_star)])]),
             'x': [(a, b, p) for (a, b), p in zip(allocator.x_star, 
                                                  allocator.p_star)]
             } 
  
    logging.info('n: %d, m: %d, d: %d, tx: %.2f, tp: %.2f, w: %.2f, r: %.2f, x: [%s]' %
                 (stats['n'], stats['m'], stats['d'], stats['tx'], stats['tp'],
                  stats['w'], stats['r'], ','.join(['(%d, %.2f, %.2f)' % (a, b, p) for (a,b,p) in stats['x']])))
    if to_output_local:
      sys.stdout.write('%d %.2f %.2f %d %d %d %.4f %.4f %.2f %.2f %s\n' %
                       (keyword, sigma, beta, stats['n'], stats['m'], stats['d'], stats['tx'], stats['tp'],
                        stats['w'], stats['r'], ' '.join(['%d %.2f %.2f' % (a, b, p) for (a,b,p) in stats['x']])))
    # Update sums
    sum_bids += sum([v for (a,v) in allocator.x_star])
    sum_welfares += allocator.w_star
    sum_prices += sum(allocator.p_star)
    sum_revenues += stats['r']
    sum_tx += stats['tx']
    sum_tp += stats['tp'] 
    
    #sum_prices += configuration['prices']
    #if use_local:
    #  sum_optimal_bids += configuration['optimal_bids']
  logging.info('b_bar: %.2f, p_bar: %.2f, w_bar: %.2f, r_bar: %.2f, tw_bar: %.4f, tp_bar: %4f' %
               (float(sum_bids) / cnt, 
                float(sum_prices) / cnt, 
                float(sum_welfares) / cnt, 
                float(sum_revenues) / cnt,
                float(sum_tx) / cnt,
                float(sum_tp) / cnt
                ))
  if to_output_aggregate:
    sys.stdout.write('%.2f %.2f %.2f %.2f %.4f %.4f' % 
                     (float(sum_bids) / cnt, 
                      float(sum_prices) / cnt, 
                      float(sum_welfares) / cnt, 
                      float(sum_revenues) / cnt,
                      float(sum_tx) / cnt,
                      float(sum_tp) / cnt))
    
def adjust_bids(keyword2bids, advertiser2conflicts, beta):
  for keyword, advertiser2bid in keyword2bids.iteritems():
    for advertiser, bid in advertiser2bid.iteritems():
      d = 0
      for conflict in advertiser2conflicts[advertiser]:
        if conflict in advertiser2bid:
          d += 1
      #d = len(advertiser2conflicts[advertiser]) 
      if d > 0:
        advertiser2bid[advertiser] = bid / ((1.0 - beta) ** d)
     

def read_bids(bids_fn):
  bids_f = open(bids_fn)
  bids_f.readline()                      # skip header
  keyword2bids = defaultdict(dict)
  for line in bids_f:
    ws = line.strip().split()
    advertiser = int(ws[0])
    keyword = int(ws[1])
    bid = float(ws[2])
    keyword2bids[keyword][advertiser] = bid
  bids_f.close()
  return keyword2bids

def read_conflicts(conflicts_fn):
  advertiser2conflicts = defaultdict(set)
  for line in open(conflicts_fn):
    ws = map(int, line.split()[0:2])
    
    advertiser2conflicts[ws[0]].add(ws[1])
    #advertiser2conflicts[ws[1]].add(ws[0])
  return advertiser2conflicts

if __name__ == '__main__':
  sys.exit(main(sys.argv))
