'''
Created on Jul 22, 2011

@author: Panagiotis
'''
import logging
from collections import defaultdict
from operator import itemgetter, attrgetter

from cvxopt import matrix, solvers, spmatrix

class Bid:
  alpha = 0.8
  
  def __init__(self, bidder, items, price, conflicts=None):
    self.bidder = bidder
    self.items = items            # set
    self.price = price
    self.weight = float(price) / (len(items) ** Bid.alpha)
    self.conflicts = conflicts
    
  def __str__(self):
    return 'items:[%s]; conflicts:[%s]; v:%.2f' \
            % (','.join(map(str, self.items)),
               ','.join(map(str, self.conflicts)),
               self.price)
            

class ConflictGraph:
  def __init__(self, alpha=0.8):
    self.v2bid = {}
    self.v2conflicts = defaultdict(set)
    self.alpha = 0.8
    self._num_conflicts = 0
    self._num_bids = 0

  def add_bid(self, j, bid, conflicts=None):
    self.v2bid[j] = bid
    if conflicts is None: conflicts = bid.conflicts
    if conflicts is not None:
      for k in conflicts:
        self.add_conflict(j, k)
    else:
      for k, bidk in self.v2bid.iteritems():
        if k != j and not bid.items.isdisjoint(bidk.items):
          self.add_conflict(j, k)
    self._num_bids += 1
            
  def add_conflict(self, j, k):
    #if u in self.v2weight and v in self.v2weight:
    if k not in self.v2conflicts[j]:
      self.v2conflicts[j].add(k)
      self._num_conflicts += 1
    if j not in self.v2conflicts[k]: 
      self.v2conflicts[k].add(j)
      self._num_conflicts += 1
  
  def get_argmax_weight(self):
    argmax_weight, max_weight = \
      max(self.v2bid.iteritems(), 
          key=lambda x:attrgetter('weight')(itemgetter(1)(x)))
    return argmax_weight

  def get_argmax_price(self):
    argmax_price, max_price = \
      max(self.v2bid.iteritems(), 
          key=lambda x:attrgetter('price')(itemgetter(1)(x)))
    return argmax_price

  def remove_bid(self, j):
    # Remove bid
    if j in self.v2bid: del self.v2bid[j]
    # Remove conflicts
    removed_conflicts = self.v2conflicts[j]
    self._num_conflicts -= len(removed_conflicts)
    if j in self.v2conflicts: del self.v2conflicts[j]
    for k in removed_conflicts:
      if k in self.v2conflicts: 
        self.v2conflicts[k].discard(j)
        self._num_conflicts -= 1
    return removed_conflicts
  
  def is_complete(self):
    return self._num_conflicts == self._num_bids * (self._num_bids - 1)
  
  def has_no_edges(self):
    return self._num_conflicts == 0
    
  def __str__(self):
    return '\n'.join(['%d, %s, %s' % (j, bid, str(self.v2conflicts[j])) 
                      for j, bid in self.v2bid.iteritems()])


class Solver:
  def __init__(self, bids):
    self.bids = list(bids)
    self.n = len(bids)
    self._init_items()
    logging.debug('Initialize conflict graph...')
    self._init_conflict_graph()
    logging.debug('...done.')
    self._init_lp_matrices()
    self._eps = 1e-6
    self.debug_cnt = 0
    
  def _init_items(self):
    self.items = set()
    for bid in self.bids:
      for item in bid.items:
        self.items.add(item)
    self.m = len(self.items)
    index = 0
    self.item2index = {}
    for item in self.items:
      self.item2index[item] = index
      index += 1
      
  def _init_conflict_graph(self):
    self.graph = ConflictGraph()
    for j, bid in enumerate(self.bids):
      self.graph.add_bid(j, bid)
      
  def _init_lp_matrices(self):
    self._lp_c = matrix([-bid.price for bid in self.bids], tc='d')
    self._lp_b = matrix([1.0] * self.m + [1.0] * self.n + [0] * self.n, tc='d')
    
    A_I = []
    A_J = []
    for j, bid in enumerate(self.bids):
      for item in bid.items:
        A_I.append(self.item2index[item])
        A_J.append(j)
    self._lp_A = spmatrix([1.0] * len(A_I) + [1.0] * self.n + [-1.0] * self.n,
                          A_I + range(self.m, self.m + self.n + self.n), 
                          A_J + range(self.n) + range(self.n), tc='d')
    self._x_lp_star = matrix(0.0, size=(self.n, 1))
    self._s_lp_star = matrix(0.0, size=(self.m + 2 * self.n, 1))

        
  def solve(self):
    self.f_star = 0
    self.x_star = []  
    self._cabob(0, [], set(self.items), 0)
    return self.f_star, self.x_star
    
  def _cabob(self, g, x, items_free, min_revenue, lp_primalstart=None):
    ''' Implementation of the paper CABOB method.
    '''
    logging.debug('CABOB: %.3f (%s)' % (g, ','.join(['%d' % (xj) for xj in x])))
    # 1
    # NO_EDGES special case
    if self.graph.has_no_edges():
      f_out = g
      for j, bid in self.graph.v2bid.iteritems():
        f_out += bid.price
      if f_out > self.f_star:
        self.x_star = x + self.graph.v2bid.keys()
        self.f_star = f_out
        logging.debug('Updated solution (NO EDGES): %.4f (%s)' % 
                      (self.f_star, ','.join(['%d' % (xj) for xj in self.x_star])))
      return f_out
    # COMPLETE special case
    if self.graph.is_complete():
      j = self.graph.get_argmax_price()
      f_out = g + self.bids[j].price
      if f_out > self.f_star:
        self.x_star = x + [j]
        self.f_star = f_out
        logging.debug('Updated solution (COMPLETE): %.4f (%s)' % 
                      (self.f_star, ','.join(['%d' % (xj) for xj in self.x_star])))
      return f_out
    # 2 (run DFS on G) - omitted
    # 3
    new_lp_primalstart = self._solve_lp(x, items_free, lp_primalstart)  
    upper = self._f_lp_star
    # 4
    if upper <= min_revenue:
      return 0
    # 5
    if self._is_integer():
      f_out = g + self._f_lp_star
      if f_out > self.f_star:
        self.f_star = f_out
        self.x_star = list(x)
        for j, xj in enumerate(self._x_lp_star):
          if abs(xj - 1) < self._eps: self.x_star.append(j)
        logging.debug('Update solution (INTEGER): %.4f (%s)' % 
                      (self.f_star, ','.join(['%d' % (xj) for xj in self.x_star])))
      return f_out
    # 6
    x_lower, f_lower, items_lower = self._calculate_lower_bound()
    # 7
    delta = g + f_lower - self.f_star
    # 8
    if delta > 0:
      self.f_star += delta
      self.x_star = x + x_lower
      min_revenue += delta
      logging.debug('Update solution (LOWER): %.4f (%s)' % 
                      (self.f_star, ','.join(['%d' % (xj) for xj in self.x_star])))
    # 9 - omitted, because graph is connected
    # 10
    k = self.graph.get_argmax_weight()
    # 10a
    conflicts_k = self.graph.remove_bid(k)
    # 10b
    conflicts = {}
    for j in conflicts_k:
      conflicts[j] = self.graph.remove_bid(j)
    # 10c
    f_star_old = self.f_star
    # 10b
    f_in = self._cabob(g + self.bids[k].price, x + [k], 
                       items_free - self.bids[k].items, 
                       min_revenue - self.bids[k].price,
                       new_lp_primalstart)
    # 10e
    min_revenue += self.f_star - f_star_old
    # 10f
    for j in conflicts_k:
      self.graph.add_bid(j, self.bids[j], conflicts[j])
    # 10g
    f_star_old = self.f_star
    # 10h
    f_out = self._cabob(g, x, items_free, min_revenue, lp_primalstart)
    # 10i
    min_revenue += self.f_star - f_star_old
    # 10j
    self.graph.add_bid(k, self.bids[k], conflicts_k)
    # 10k
    return max(f_in, f_out)
      
  def _is_integer(self):
    for xj in self._x_lp_star:
      if abs(xj) > self._eps and abs(xj - 1) > self._eps:
        return False
    return True

  def _solve_lp(self, x, items_free, lp_primalstart=None):
    ''' Solves LP equivalent and stores solution on _x_lp_star, _s_lp_star and _f_lp_star.
    '''
    bid_indexes = self.graph.v2bid.keys()     # Active bids
    item_indexes = [self.item2index[item] for item in items_free]
    rows = item_indexes + map(lambda i : i + self.m, bid_indexes) +  \
                          map(lambda i : i + self.m + self.n, bid_indexes)
    lp_primalstart_reduced = None
    if lp_primalstart is not None:
      lp_primalstart_reduced = {'x': lp_primalstart['x'][bid_indexes],
                                's': lp_primalstart['s'][rows]}
      
    sol = solvers.lp(self._lp_c[bid_indexes],
                     self._lp_A[rows, bid_indexes],
                     self._lp_b[rows],
                     primalstart=lp_primalstart_reduced)
    # Update solution
    self._x_lp_star[:] = 0.0
    self._x_lp_star[bid_indexes] = sol['x']
    self._s_lp_star[:] = 0.0
    self._s_lp_star[rows] = sol['s']
    self._f_lp_star = -sol['primal objective']
    # Return new lp_primalstart
    return {'x': matrix(self._x_lp_star),
            's': matrix(self._s_lp_star)
            }
  
  def _calculate_lower_bound(self):
    x_lower = []
    f_lower = 0
    items_lower = set()
    # Insert all bids with x_j > 0.5 in the lower bound solution
    cnt = 0  # counts inserted bids
    for j, xj in enumerate(self._x_lp_star):
      if xj > 0.5: 
        x_lower.append(j)
        f_lower += self.bids[j].price
        items_lower |= self.bids[j].items
        cnt += 1
    # Try to insert the rest of the bids in decreasing order of x_j
    for j, xj in sorted(enumerate(self._x_lp_star), 
                        key=itemgetter(1), reverse=True)[cnt:]:
      #if xj == 0:
      #  break
      if j not in self.graph.v2bid:
        continue
      if items_lower.isdisjoint(self.bids[j].items):
        x_lower.append(j)
        f_lower += self.bids[j].price
        items_lower |= self.bids[j].items
    return x_lower, f_lower, items_lower

if __name__ == '__main__':
  bids = [Bid(1, set([1,2]), 3), Bid(2, set([1,3,4]), 5), Bid(3, set([1,2,3]), 6)]
  solver = Solver(bids)
  res = solver.solve()
  print res
  bids = [Bid(1, set([1,2]), 3), Bid(2, set([1,3,4]), 5), Bid(3, set([1,2,3]), 5)]
  solver = Solver(bids)
  res = solver.solve()
  print res
