'''
Created on Jul 25, 2011

@author: Panagiotis
'''
import logging
import math

from collections import defaultdict
from datetime import datetime
from operator import itemgetter

from determination.allocator import Allocator
from cabob.solver import Bid, Solver


class CombinatorialAuction(Allocator):
  
  def __init__(self, advertiser2bid, advertiser2conflicts, rhos, keyword=None):
    Allocator.__init__(self, advertiser2bid, advertiser2conflicts, rhos, keyword)
    self._init_bids()
    self._solver = Solver(self._bids)
    
  def _init_bids(self):
    # Start with one item per slot and one for each advertiser
    max_item_id = self.m + self.n
    # Assign indexes to advertisers
    aj2j = {}
    for j, aj in enumerate(self.advertiser2bid.iterkeys()):
      aj2j[aj] = j
    # Find all bids on the i-th slot
    item_bids = []
    for i in range(len(self.rhos)):
      item_bids.append(set([j * self.m + i for j in aj2j.itervalues()]))
    # Create bids with dummy items for conflicts
    conflict2item = defaultdict(int)
    self._bids = []
    for aj, j in sorted(aj2j.iteritems(), key=itemgetter(1)):
      items_j = set([self.m + j])
      # introduce conflicts among all advertiser j bids
      bid_conflicts = set(range(j * self.m, (j + 1) * self.m))
      # introduce bid conflicts for advertiser conflicts
      conflict_cnt = 0
      for ak in self.advertiser2conflicts[aj]:
        # Ignore conflicts if ak does not participate in the auction
        if ak not in self.advertiser2bid:
          continue
        conflict_cnt += 1
        conflict = (min(aj, ak), max(aj, ak))
        if conflict not in conflict2item:
          conflict2item[conflict] = max_item_id
          max_item_id += 1
        items_j.add(conflict2item[conflict])
        k = aj2j[ak]
        bid_conflicts.update(range(k * self.m, (k + 1) * self.m))
      # introduce conflicts with all bids that have the same real slot (done below)
      # Append new bids
      for i, rho in enumerate(self.rhos):
        self._bids.append(Bid(aj, set([i]) | items_j, 
                              self.advertiser2bid[aj] * rho, 
                              conflicts=(bid_conflicts | item_bids[i]) - set([j * self.m + i])))
      num_conflicts = self.m - 1 + conflict_cnt * (self.m - 1) + self.n - 1
      if len(self._bids[-1].conflicts) != num_conflicts:
        pass
      

  def allocate_slots(self, A=None, to_price=False):
    t_start = datetime.now()
    self._solver.solve()
    t_end = datetime.now()
    if to_price:
      self.p_star = [-1] * len(self._solver.x_star)
      logging.info("Pricing is not supported.")
    # Update members
    self.delta_winners = (t_end - t_start).total_seconds()
    self.delta_price = 0
    self.x_star = []
    self.w_star = 0.0
    for j in self._solver.x_star:
      aj = self._bids[j].bidder
      uj = self._bids[j].price
      vj = self.advertiser2bid[aj]
      self.x_star.append((aj, vj))
      self.w_star += uj
    self.x_star.sort(key=itemgetter(1), reverse=True)
    if abs(self.w_star - self._solver.f_star) > 1e-4:
      logging.warn('LP solution and welfare differ (%d)' % self.keyword)
      #lp_winners = [(self._bids[j].bidder, 
      #               self.advertiser2bid[self._bids[j].bidder], 
      #               self._bids[j].price) 
      #              for j in self._solver.x_star]
      #lp_winners.sort(key=itemgetter(1), reverse=True)
      #rho2i = {math.floor(10000 * rho): i for i, rho in enumerate(self.rhos)}
      #logging.warn('LP solution and welfare differ (%d):\n%.3f (%s)\n%.3f (%s)' 
      #             % (self.keyword,
      #                self._solver.f_star, ', '.join(['%d:%.2f:%.2f:%d' % (aj, vj, uj, rho2i[math.floor(10000 * uj / vj)]) 
      #                                               for (aj, vj, uj) in lp_winners]),
      #                self.w_star, ', '.join(['%d:%.2f:%.2f:%d' % (aj, vj, vj*rho, i) 
      #                                       for i, ((aj, vj), rho) in enumerate(zip(self.x_star, self.rhos[:len(self.x_star)]))])))
    
if __name__ == '__main__':
  logging.basicConfig(format="%(asctime)s\t%(levelname)s\t%(message)s", level=logging.DEBUG)
  advertiser2bid = {1:10, 2:8, 3:5, 4:2}
  advertiser2conflicts = {1:set([2,3])}
  allocator = CombinatorialAuction(advertiser2bid, advertiser2conflicts, [0.5, 0.25])
  print '\n'.join([str(j) + ' ' + str(bid) for j, bid in enumerate(allocator._bids)])
  allocator.allocate_slots()
  print allocator.x_star, allocator.w_star