'''
Created on Apr 19, 2011

@author: Panagiotis
'''

import logging
import heapq

from itertools import chain, combinations
from collections import defaultdict
from copy import deepcopy


class ConflictGraph:
  def __init__(self, greedy=1):
    self.v2weight = defaultdict(float)
    self.v2conflicts = defaultdict(set)
    self._max_node_id = 0
    if greedy == 1:
      self.normalization_factor = self.degree_normalization_factor
    else:
      self.normalization_factor = self.neighborhood_normalization_factor
    
    
  def add_node(self, v, weight):
    self.v2weight[v] = weight
    if v > self._max_node_id:
      self._max_node_id = v
  
  def add_spare_node(self):
    self.add_node(self._max_node_id + 1, 0)
    
  def add_edge(self, u, v):
    #if u in self.v2weight and v in self.v2weight:
    self.v2conflicts[u].add(v)
    self.v2conflicts[v].add(u)
    
  def remove_node(self, v):
    if v in self.v2conflicts:
      for u in self.v2conflicts[v]:
        self.v2conflicts[u].remove(v)
      del self.v2conflicts[v]
    
    if v in self.v2weight:
      del self.v2weight[v]

  def num_nodes(self):
    return len(self.v2weight)
  
  def get_node_weight(self, v):
    return self.v2weight[v]
  
  def get_neighbors(self, v):
    return self.v2conflicts[v]
  
  def complement(self, nodes):
    neighborhood = set(nodes)
    for neighbors in self.v2conflicts.itervalues():
      neighborhood |= neighbors
    complement = set()
    for v in self.v2weight.iterkeys():
      if v not in neighborhood:
        complement.add(v)
    return complement
  
  def degree_normalization_factor(self, v):
    return float(len(self.v2conflicts[v]) + 1)
  
  def neighborhood_normalization_factor(self, v):
    v_weight = self.v2weight[v]
    if v_weight == 0:
      return 1.0
    else:
      return float(v_weight + sum([self.v2weight.get(u, 0) for u in self.v2conflicts[v]]))
  
  def next_winner(self, reserve_price=0.01):
    top_norm_weights = []
    for v, weight in self.v2weight.iteritems():
      norm_weight = weight / self.normalization_factor(v)
      if len(top_norm_weights) < 2:
        heapq.heappush(top_norm_weights, (norm_weight, weight, v))
      else:
        heapq.heappushpop(top_norm_weights, (norm_weight, weight, v))
    # Pick winner and set prices
    top_norm_weights.reverse()
    if len(top_norm_weights) > 0:
      norm_winner_weight, weight, winner = top_norm_weights[0]
      second_price = reserve_price
      multiplier = self.normalization_factor(winner)
      if len(top_norm_weights) > 1:
        norm_second_weight, second_weight, second = top_norm_weights[1]
        if norm_second_weight > second_price:
          second_price = norm_second_weight
        if norm_winner_weight == second_price:
          second_price = second_weight
          multiplier = 1
      bid = self.get_node_weight(winner)
      price = second_price * multiplier
      return winner, bid, price    
    else:
      raise NoWinnerError()
    

class NoWinnerError(Exception):
  def __init__(self):
    pass
  

def determine_winners(advertiser2bid, advertiser2conflicts, k=10, use_local=True, greedy=1):
  Gc = ConflictGraph(greedy=greedy)
  for advertiser, bid in advertiser2bid.iteritems():
    Gc.add_node(advertiser, bid)
    if advertiser in advertiser2conflicts:
      for conflict in advertiser2conflicts[advertiser]:
        Gc.add_edge(advertiser, conflict)
  # Add zero nodes
  for i in range(k-1):
    Gc.add_spare_node()
  
  winners = []
  winners_bids = 0
  winners_prices = 0
  G = deepcopy(Gc)
  while k > 0 and G.num_nodes() > 0:
    winner, bid, price = G.next_winner()
    logging.debug('Adding %d with weight %.2f. Price: %.2f' % (winner, bid, price))
    winners.append((winner, price))
    winners_bids += bid
    winners_prices += price
    
    neighbors = G.get_neighbors(winner)
    for u in chain([winner], neighbors):
      G.remove_node(u)
    k -= 1
  logging.debug('Sum = %.2f, (%s)' % 
                (winners_bids,
                 ','.join(['%d:%.2f' % (winner, price) 
                           for winner, price in winners])))
  results = {'winners': winners,
             'bids': winners_bids,
             'prices': winners_prices}
  if use_local:
    optimal_winners, optimal_bids = \
    optimize_bids(set([advertiser for (advertiser, price) in winners]), Gc, k)
    results.update({'optimal_winners': optimal_winners,
                    'optimal_bids': optimal_bids})
  return results


def optimize_bids(seed, Gc, k=10):
  # Local Search
  seed_bids = sum([Gc.get_node_weight(v) for v in seed])
  candidates_pool = Gc.complement(seed)
  is_improved = False
  while 1:  
    best_improvement = 0
    replaced_pair = set()
    new_pair = set()
      
    for u, v in combinations(seed, 2):  
      pair = set([u, v])
      pair_bids = Gc.get_node_weight(u) + Gc.get_node_weight(v)

      current_winners = seed - pair
      current_bids = seed_bids - pair_bids

      candidates = []
      for candidate in candidates_pool | Gc.v2conflicts[u] | Gc.v2conflicts[u]:
        if len(Gc.v2conflicts[candidate] & current_winners) == 0:
          candidate_bid = Gc.v2weight[candidate]
          if len(candidates) < 2:
            heapq.heappush(candidates, (candidate_bid, candidate))
          else:
            heapq.heappushpop(candidates, (candidate_bid, candidate)) 
      # Check for improvement
      candidate_bids = sum([bid for (bid, candidate) in candidates])
      logging.debug('%d,%d (%.2f) -> %s (%.2f)' % (u, v, pair_bids, 
                                                   ','.join([str(y) for (x,y) in candidates]),
                                                   candidate_bids))
      improvement = candidate_bids - pair_bids
      if improvement > best_improvement:
        logging.debug('Improvement: %f' % (improvement))
        is_improved = True
        best_improvement = improvement
        replaced_pair = pair
        new_pair = set([candidate for (bid, candidate) in candidates])
    if not is_improved:
      break
    else:
      seed = (seed - replaced_pair) | new_pair
      candidates_pool = Gc.complement(seed) # TODO: optimize update 
      seed_bids = sum([Gc.get_node_weight(advertiser) for advertiser in seed])
      is_improved = False
      logging.debug('Sum = %.2f, (%s)' % (seed_bids, ','.join([str(winner) for winner in seed])))
  return seed, seed_bids  

# Test
if __name__ == '__main__':
  logging.basicConfig(format="%(asctime)s\t%(levelname)s\t%(message)s", level=logging.DEBUG)
  graph = ConflictGraph()
  advertiser2bid = {1:10, 2:8, 3:5, 4:2}
  advertiser2conflicts = {1:set([2,3])}
  #advertiser2bid = {1:10, 2:8, 3:6, 4:5, 5:2}
  #advertiser2conflicts = {1:set([2,3]), 2:set([5])}
  # Weights
  graph.add_node(1, 10)
  graph.add_node(2, 8)
  graph.add_node(3, 6)
  graph.add_node(4, 5)
  graph.add_node(5, 2)
  # Conflicts
  graph.add_edge(1, 2)
  graph.add_edge(1, 3)
  graph.add_edge(2, 5)
  
  configuration = determine_winners(advertiser2bid, advertiser2conflicts, k=3, use_local=True)
  print configuration
  configuration = determine_winners(advertiser2bid, advertiser2conflicts, k=3, use_local=True, greedy=2)
  print configuration

  